from marshmallow import Schema, fields


class BaseCookie:
    """
    The BaseCookie represents the
    a customizable cookie to control
    updates metadata on devices.
        :::param cookie
    """

    def __init__(self, cookie) -> 'BaseCookie':
        self.cookie = cookie

class CookieSchema(Schema):
    """
    The CookieSchema handles the
    serialization of the BaseCookie
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    cookie = fields.Field(attribute='cookie')



class BaseDevice:
    """
    The BaseDevice class represents
    the basic metadata of a device
    integrations.
        :::param external_device_id
        :::param friendly_name
        :::param device_handler_type
        :::param device_cookie
        :::param unique_device_id
    """

    def __init__(self, external_device_id: str, friendly_name: str, device_unique_id: str, device_cookie: str, device_handler_type: str) -> 'BaseDevice':
        self.external_device_id = external_device_id
        self.friendly_name = friendly_name
        self.device_unique_id = device_unique_id
        self.device_cookie = device_cookie
        self.device_handler_type = device_handler_type



from typing import List
from marshmallow import Schema, fields


class DeviceContext:
    """
    The DeviceContext represents the
    context/location metadata of the
    device integration.
        :::param room_name
        :::param groups
        :::param categories
    """

    def __init__(self, room_name: str, groups: List[str], categories: List[str]) -> 'DeviceContext':
        self.room_name = room_name
        self.groups = groups or list()
        self.categories = categories or list()

class DeviceContextSchema(Schema):
    """
    The DeviceContextSchema handles the
    serialization of the DeviceContext
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    roomName = fields.Str(attribute='room_name')
    groups = fields.List(fields.Str(attribute='groups'))
    categories = fields.List(fields.Str(attribute='categories'))



class ErrorHandler:
    """
    ErrorHandler class to pretty print
    the different exceptions raised through
    development stages.

    Handlers supported:
        - not_implemented_error
    """

    def not_implemented_error(self, origin: str):
        error_msg = '[%s] - Interaction resource handler not implemented' % origin
        raise NotImplementedError(error_msg)



from marshmallow import Schema, fields
from stschema.util.base_modules import CookieSchema, StateSchema


class DeviceStateSchema(Schema):
    """
    The DeviceStateSchema handles the
    serialization of the SchemaDevice
    class supporting the nested
    CookieSchema and StateSchema.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    externalDeviceId = fields.Str(attribute='external_device_id')
    deviceCookie = fields.Nested(CookieSchema, attribute='device_cookie')
    states = fields.List(fields.Nested(StateSchema), attribute='states')



from marshmallow import Schema, fields
from stschema.util.base_modules import BaseErrorSchema


class DeviceErrorSchema(Schema):
    """
    The DeviceErrorSchema handles the
    serialization of the SchemaDevice
    class supporting the nested
    BaseErrorSchema.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    externalDeviceId = fields.Str(attribute='external_device_id')
    deviceError = fields.List(fields.Nested(BaseErrorSchema), attribute='device_error')



from marshmallow import Schema, fields, pre_dump


class CallbackAuthentication:
    """
    The CallbackAuthentication class
    represents the base information
    to authenticate the third-party cloud
    and perform the following interaction
    types:
        - Access Token Request.
        - Refresh Access Tokens.
        - State Callback.
        - Discovery Callback.
    """
    def __init__(self, client_id: str, client_secret: str, code: str, refresh_token: str, grant_type: str) -> 'CallbackAuthentication':
        self.client_id = client_id
        self.client_secret = client_secret
        self.code = code
        self.refresh_token = refresh_token
        self.grant_type = grant_type

class CallbackAuthSchema(Schema):
    """
    The CallbackAuthSchema handles the
    serialization of CallbackAuthentication
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    @pre_dump
    def _verify_dump(self, data, **kwargs):
        self.dump_fields.update(
            clientId = fields.Str(attribute='client_id', required=True, dump_only=True),
            clientSecret = fields.Str(attribute='client_secret', required=True, dump_only=True),
            grantType = fields.Str(attribute='grant_type', required=True, dump_only=True))
        if data.code:
            self.dump_fields.update(
                code = fields.Str(attribute='code', required=True, dump_only=True))
        elif data.refresh_token:
            self.dump_fields.update(
                refreshToken = fields.Str(attribute='refresh_token', required=True, dump_only=True))
        return data




from stschema.util.base_modules import BaseHeaders


class BaseResponse:
    """
    The BaseResponse represents the
    base information that will be extended
    in response to the following interaction
    types.
        - discoveryRequest
        - stateRefreshResponse
        - commandRequest
        - discoveryCallback
        - stateCallback
        - accessTokenRequest
    """
    # Future implementation reference
    devices = list()
    device_state = list()
    authentication = dict()
    global_error = dict()

    def __init__(self, interaction_type: str, request_id: str) -> 'BaseResponse':
        self.headers = BaseHeaders(interaction_type=interaction_type, request_id=request_id)



from .base_device import *
from .base_headers import *
from .base_response import *
from .base_state import *
from .device_context import *
from .device_cookie import *
from .base_error import *
from .error_enum import *
from .manufacturer_info import *
from .callback_authentication import *
from .base_authentication import *



from marshmallow import Schema, fields, pre_dump
from stschema.schema_device.schemas import DeviceDiscoverySchema
from stschema.util.base_modules import BaseResponse, HeadersSchema


class DiscoveryResponse(BaseResponse):
    """
    The DiscoveryResponse is an object
    representation in response to the
    "discoveryRequest" Interaction Type.
    Inherits from BaseResponse.

        :::param devices
        :::param request_id
    """

    def __init__(self, devices: list, request_id: str, interaction_type: str='discoveryResponse') -> 'DiscoveryResponse':
        BaseResponse.__init__(self, interaction_type=interaction_type, request_id=request_id)
        self.devices = devices


class DiscoveryResponseSchema(Schema):
    """
    The DiscoveryResponseSchema handles
    the serialization of the DiscoveryResponse
    class supporting the nested
    DeviceDiscoverySchema, HeadersSchema,
    and AuthenticationSchema.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    @pre_dump
    def _verify_dump(self, data, **kwargs):
        # Declare required attributes
        self.dump_fields.update(
            headers = fields.Nested(HeadersSchema, attribute='headers', required=True),
            devices = fields.List(fields.Nested(DeviceDiscoverySchema, attribute='devices', required=True)))
        # Verify Authentication non-required attribute.
        if data.authentication:
            self.dump_fields.update(
                authentication = fields.Nested(AuthenticationSchema, attribute='authentication'))
        return data



from marshmallow import Schema, fields


class ManufacturerInfo:
    """
    The ManufacturerInfo represents
    the device's metadata in context of
    a product.
        :::param manufacturer_name
        :::param model_name
        :::param hw_version
        :::param sw_version
    """

    def __init__(self, manufacturer_name: str, model_name: str, hw_version: str, sw_version: str) -> 'ManufacturerInfo':
        self.manufacturer_name = manufacturer_name
        self.model_name = model_name
        self.hw_version = hw_version
        self.sw_version = sw_version

class ManufacturerSchema(Schema):
    """
    The ManufacturerSchema handles the
    serialization of the ManufacturerInfo
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    manufacturerName = fields.Str(attribute='manufacturer_name', required=True)
    modelName = fields.Str(attribute='model_name', required=True)
    hwVersion = fields.Str(attribute='hw_version')
    swVersion = fields.Str(attribute='sw_version')



from marshmallow import Schema, fields


class BaseAuthentication:
    """
    The BaseAuthentication class represents
    the attribute "authentication" that includes
    the Authorization Token needed to perform
    successful Calbacks into the SmartThings API.
    """
    def __init__(self, token: str, token_type: str='Bearer') -> 'BaseAuthentication':
        self.token = token
        self.token_type = token_type

class AuthenticationSchema(Schema):
    """
    The AuthenticationSchema handles the
    serialization of BaseAuthentication
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """
    token = fields.Str(required=True)
    tokenType = fields.Str(required=True, attribute='token_type')



from stschema.schema_response.responses import DiscoveryResponse
from stschema.util.base_modules import BaseAuthentication


class DiscoveryCallback(DiscoveryResponse):
    def __init__(self, access_token: str, request_id: str, devices: list) -> 'DiscoveryCallback':
        DiscoveryResponse.__init__(self,
                                   devices=devices,
                                   request_id=request_id,
                                   interaction_type='discoveryCallback')



from marshmallow import Schema, fields


class BaseError:
    """
    The BaseError class represents
    an error state definition. Used as
    a DeviceError or Global Error
    instance.
        :::param error_enum
        :::param detail
    """

    def __init__(self, error_enum: str, detail: str) -> 'BaseError':
        self.error_enum = error_enum
        self.detail = detail

class BaseErrorSchema(Schema):
    """
    The BaseError handles the
    serialization of the DeviceError
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    errorEnum = fields.Str(attribute='error_enum')
    detail = fields.Str()



from stschema.schema_device import SchemaDevice
from stschema.util.base_modules import GlobalErrorEnum, BaseError
from stschema.schema_response.responses import (DiscoveryResponse,
                                                DiscoveryResponseSchema,
                                                StateResponse,
                                                StateRefreshResponseSchema,
                                                GlobalErrorResponse,
                                                GlobalErrorSchema)


class SchemaResponse:
    """
    The SchemaResponse class handles
    the instantiation and serialization
    of the SchemaConnector JSON Responses.
    """

    @classmethod
    def discovery_response(cls, devices: list, request_id: str):
        """
        The discovery_response classmethod will handle
        the instantiation and serialization of the
        DiscoveryResponse into a readable JSON object
        response.
            :::param devices: list of Device instances
            :::param request_id: str
        """
        return cls._validate_schema_response(devices, request_id, cls._discovery_response)

    @classmethod
    def state_refresh_response(cls, devices: list, request_id: str):
        """
        The state_refresh_response classmethod will
        handle the instantiation and serialization of the
        StateRefreshResponse into a readable JSON object
        response.
            :::param devices: list of Device instances
            :::param request_id: str
        """
        return cls._validate_schema_response(devices, request_id, cls._state_refresh_response)

    @classmethod
    def command_response(cls, devices:list, request_id: str):
        """
        The command_response classmethod will handle
        the instantitation and serialization of a new
        State isntance in response of a Command Request.
            :::param devices: list of Device instances
            :::param request_id: str
        """
        return cls._validate_schema_response(devices, request_id, cls._command_response)

    @staticmethod
    def _validate_schema_response(devices: list, request_id: str, response_callback):
        # This private classmethod will validate the data passed
        # to instantiate a DiscoveryResponse or StateRefreshResponse
        # (CommmandResponse).
        # It will enhance early development stages of Schema Connector
        # instances.
        if not isinstance(devices, list):
            raise TypeError('"devices" argument must be instance of list not %s' % type(devices))
        if devices:
            for device in devices:
                if not isinstance(device, SchemaDevice):
                    raise TypeError('"devices" items must be instances of %s, not %s' % (SchemaDevice, device))
        # Validate data type of request_id.
        # It must be string value
        if not isinstance(request_id, str):
            raise TypeError('request_id must be instance of str not %s' % type(request_id))
        return response_callback(devices, request_id)

    @staticmethod
    def _discovery_response(devices: list, request_id: str):
        # Private method with the responsibility to
        # instantiate a DiscoveryResponse and serialize it
        # into a readable JSON object.

        response = DiscoveryResponse(devices=devices, request_id=request_id)
        discovery_schema = DiscoveryResponseSchema()
        return discovery_schema.dump(response)

    @staticmethod
    def _state_refresh_response(devices: list, request_id: str):
        # Private method with the responsibility to
        # instantiate a StateResponse and serialize it
        # into a readable JSON object.

        response = StateResponse(devices=devices, request_id=request_id)
        state_schema = StateRefreshResponseSchema()
        return state_schema.dump(response)

    @staticmethod
    def _command_response(devices: list, request_id: str):
        # Private method with the responsibility to
        # instantiate a command response(StateResponse)
        # and serialize it into a readable JSON object.

        response = StateResponse(devices=devices,
                                 request_id=request_id,
                                 interaction_type='commandResponse')
        state_schema = StateRefreshResponseSchema()
        return state_schema.dump(response)

    @classmethod
    def global_error_response(cls,
        interaction_type: str,
        request_id: str,
        error_enum: str = 'BAD-REQUEST',
        detail: str = 'invalid request arguments.'):
        """
        Defines a global error of communication
        between Clouds.
        Supported Global Error enumerators:
            - BAD-REQUEST
            - INTEGRATION-DELETED
            - INVALID-CLIENT
            - INVALID-CLIENT-SECRET
            - INVALID-CODE
            - INVALID-INTERACTION-TYPE
            - INVALID-TOKEN
            - TOKEN-EXPIRED
            - UNSUPPORTED-GRANT-TYPE

            :::param interaction_type (required)
            :::param request_id (required)
            :::param error_enum: "BAD-REQUEST" by default
            :::param detail: message or detail
            about global error.
        """
        try:
            error_enum = GlobalErrorEnum(error_enum)
        except ValueError as e:
            raise ValueError('Device error enumerator not supported: %s' % error_enum)
        else:
            if not request_id:
                raise TypeError('"request_id" argument is missing')
            else:
                global_error = BaseError(error_enum.value, detail)
                return cls._global_error_response(interaction_type,
                                                  request_id,
                                                  global_error)

    @staticmethod
    def _global_error_response(*error_info):
        response = GlobalErrorResponse(error_info[0],
                                       error_info[1],
                                       error_info[2])
        error_schema = GlobalErrorSchema()
        return error_schema.dump(response)



from marshmallow import Schema, fields
from stschema.util.base_modules import (
    HeadersSchema,
    BaseResponse,
    BaseError,
    BaseErrorSchema
)


class GlobalErrorResponse(BaseResponse):
    """
    The GlobalErrorResponse is an object
    representation of a global-contextual
    error.

        :::param interaction_type (required)
        :::param request_id (required)
        :::param global_error: BaseError instance.
    """
    def __init__(self, interaction_type: str, request_id: str, global_error: object) -> 'GlobalErrorResponse':
        BaseResponse.__init__(self, interaction_type, request_id)
        # BaseError instance
        self.global_error = global_error


class GlobalErrorSchema(Schema):
    """
    The GlobalErrorSchema handles
    the serialization of the
    GlobalErrorResponse classsupporting
    the nested BaseErrorSchema and
    HeadersSchema.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """
    headers = fields.Nested(HeadersSchema, attribute='headers')
    globalError = fields.Nested(BaseErrorSchema, attribute='global_error')



from stschema.schema_response.responses import StateResponse
from stschema.util.base_modules import BaseAuthentication

class StateCallback(StateResponse):
    """
    The StateCallback class represents
    the information used to instance
    Device State Callbacks.
    """
    def __init__(self, access_token: str, request_id: str, devices: list) -> 'StateCallback':
        StateResponse.__init__(self,
                               devices=devices,
                               request_id=request_id,
                               interaction_type='stateCallback')
        self.authentication = BaseAuthentication(access_token)



"""
The Device interface provides a series of
methods to collect the information about
the devices that will be served to the
SmartThings Schema Connector integration
type.
"""
from stschema.util.base_modules import (
    BaseDevice,
    DeviceContext,
    ManufacturerInfo,
    BaseState,
    BaseError,
    BaseCookie,
    StateErrorEnum
)


class SchemaDevice(BaseDevice):
    """
    The Device interface inherits its
    attributes from the BaseDevice class.
        :::param external_device_id (required)
        :::param friendly_name (required)
        :::param device_handler_type (required)
        :::param device_unique_id
        :::param device_cookie
    """

    def __init__(self,
                 external_device_id=None,
                 friendly_name=None,
                 device_handler_type=None,
                 device_unique_id=None,
                 device_cookie=None) -> 'SchemaDevice':
        BaseDevice.__init__(self,
                            external_device_id,
                            friendly_name,
                            device_unique_id,
                            BaseCookie(device_cookie),
                            device_handler_type)

        self.states = []
        self.device_error = None
        self.device_context = None
        self.manufacturer_info = None

    # def set_mn(self, *args_info, **kwargs_info):
    def set_mn(self, manufacturer_name, model_name, hw_version, sw_version):
        """Defines the device's manufacturer information.
        Returns an instance of the ManufacturerInfo class.
            :::param manufacturer_name (required)
            :::param model_name (required)
            :::param hw_version
            :::param sw_version"""
        self.manufacturer_info = ManufacturerInfo(
            manufacturer_name,
            model_name,
            hw_version,
            sw_version
        )

    def set_context(self, room_name, groups, categories):
        """Defines the device's context information.
        Returns an instance of the DeviceContext class.
            :::param room_name
            :::param groups
            :::param categories"""
        # Type handling for list arguments
        if not isinstance(groups, list):
            raise TypeError('"groups" must be list instance, not (%s, %s)' % (type(groups), groups))
        elif not isinstance(categories, list):
            raise TypeError('"categories" must be list instance, not (%s, %s)' % (type(categories), categories))

        self.device_context = DeviceContext(room_name,
                                            groups,
                                            categories)

    # def set_state(self, *args_info, **kwargs_info):
    def set_state(self, capability, attribute, value, unit=None, component='main'):
        """Defines the device's state by
        adding one capability-state per call.
        Returns an instance of the BaseState class."""
        new_state = BaseState(capability=capability,
                              attribute=attribute,
                              value=value,
                              unit=unit,
                              component=component)
        self.states.append(new_state)

    def set_error_state(self, error_enum: str='DEVICE-UNAVAILABLE', detail: str='unexpected error occurred.'):
        """
        Defines the error state of the device.
        Supported device error enumerators:
            - DEVICE-UNAVAILABLE
            - CAPABILITY-NOT-SUPPORTED
            - RESOURCE-CONSTRAINT-VIOLATION
            - DEVICE-DELETED

            :::param error_enum: "DEVICE-UNAVAILABLE" by default.
            :::param detail: detail or message about device error
        """

        try:
            error_enum = StateErrorEnum(error_enum)
        except ValueError as e:
            raise ValueError('Device error enumerator not supported: %s' % error_enum)
        else:
            # self._set_error_state(error_enum.value, detail)
            err = BaseError(error_enum=error_enum.value, detail=detail)
            self.device_error = [err]
